#include <iostream>

using namespace std;

#define LF {cout << "\n";}

class String
{
public:
  //
  // constructor
  //
  String()                               { cout << this << "::constructor() "; }
  String(char* p)                        { cout << this << "::constructor(char*) "; }
  String(const char* p)                  { cout << this << "::constructor(const char*) "; }
  String(const String& rhs)              { cout << this << "::constructor(const " << &rhs << "::String&) "; }

  //
  // destructor
  //
  virtual ~String()                      { cout << this << "::destructor "; }

  //
  // operator
  //
  String& operator = (char* p)           { cout << this << "::operator=(char*) ";        return *this; }
  String& operator = (const char* p)     { cout << this << "::operator=(const char*) ";  return *this; }
  String& operator = (const String& rhs) { cout << this << "::operator=(const " << &rhs << "::String) ";  return *this; }
};

String g_s;

//
// param_test
//
void pointer_param_test(String* s)
{
}

void reference_param_test(String& s)
{
}

void object_param_test(String s)
{
}

//
// return_test
//
String* pointer_return_test()
{
  return &g_s;
}

String& reference_return_test()
{
  return g_s;
}

String object_return_test()
{
  return g_s;
}

void basic_test()
{
  {
    cout << "\nString s(\"test\");\t";                  String s("test");
    LF;
  } LF;

  {
    cout << "\nString s = \"test\";\t";                 String s = "test";
    LF;
  } LF;

  {
    cout << "\nString s;\t";                            String s;
    cout << "\ns = \"test\";\t";                        s = "test";
    LF;
  } LF;

  {
    cout << "\nString s(g_s);\t";                       String s(g_s);
    LF;
  } LF;

  {
    cout << "\nString s = g_s;\t";                      String s = g_s;
    LF;
  } LF;

  {
    cout << "\nString s;\t";                            String s;
    cout << "\ns = g_s;\t";                             s = g_s;
    LF;
  } LF;

  {
    cout << "\nconst char* p = \"test\";\t";            const char* p = "test";
    cout << "\nString s(p);\t";                         String s(p);
    LF;
  } LF;

  {
    cout << "\nchar* p = \"test\";\t";                  char* p = "test";
    cout << "\nString s(p);\t";                         String s(p);
    LF;
  } LF;
}

void param_test()
{
  {
    cout << "\npointer_param_test(&g_s);\t";            pointer_param_test(&g_s);
    LF;
  } LF;

  {
    cout << "\nreference_param_test(g_s);\t";           reference_param_test(g_s);
    LF;
  } LF;

  {
    cout << "\nobject_param_test(g_s);\t";              object_param_test(g_s);
    LF;
  } LF;
}

void return_test()
{
  {
    cout << "\npointer_return_test();\t";               pointer_return_test();
    LF;
  } LF;

  {
    cout << "\nString* s;\t";                           String* s;
    cout << "\ns = pointer_return_test();\t";           s = pointer_return_test();
    LF;
  } LF;

  {
    // String& s; // compile error. reference variable must be initialized with its declaraion.
  }

  {
    cout << "\nString& s = reference_return_test();\t"; String& s = reference_return_test();
    LF;
  } LF;

  {
    cout << "\nString s = reference_return_test();\t";  String s = reference_return_test();
    LF;
  } LF;

  {
    cout << "\nString s;\t";                            String s;
    cout << "\ns = reference_return_test();\t";         s = reference_return_test();
    LF;
  } LF;

  {
    cout << "\nobject_return_test();\t";                object_return_test();
    LF;
  } LF;

  {
    cout << "\nString s(object_return_test());\t";      String s(object_return_test());
    LF;
  } LF;

  {
    cout << "\nString s = object_return_test();\t";     String s = object_return_test();
    LF;
  } LF;

  {
    cout << "\nString s;\t";                            String s;
    cout << "\ns = object_return_test();\t";            s = object_return_test();
    LF;
  }
}

int main()
{
  LF;LF;LF;
  basic_test();
  param_test();
  return_test();
  LF;LF;LF;
  return 0;
}
